package pathgame.tilemap;

import java.util.function.Function;
import pathgame.gameplay.Player;

/** Base class for tiles. Tiles are registered on construction.
 *
 * @author kajetan
 */
public class Tile
{

    /** A builder for tiles.
     *
     */
    public static class TileBuilder
    {
        private int energyCost = 2;
        private float forestReplaceChance = 1.0f;
        private Function<Player, Integer> speedUp = (p) -> 0;
        private boolean canHostTown = true;
        private boolean blocksMovement = false;
        private TileType type = TileType.LAND;
        private boolean canHostPath = false;
        private boolean path = false;
        private TileRenderType renderType = TileRenderType.NORMAL;
        
        private TileBuilder()
        {
        }
        
        /** Returns a new tile builder;
         *
         * @return a new tile builder
         */
        public static TileBuilder create()
        {
            return new TileBuilder();
        }

        /** Sets the base energy cost of the tile.
         *
         * @param energyCost the base energy cost of the tile.
         * @return the tile builder
         */
        public TileBuilder setEnergyCost(int energyCost)
        {
            this.energyCost = energyCost;
            return this;
        }

        /** Sets the chance that the tile can be replaced by forest.
         *
         * @param forestReplaceChance the chance that the tile can be replaced by forest.
         * @return the tile builder
         */
        public TileBuilder setForestReplaceChance(float forestReplaceChance)
        {
            this.forestReplaceChance = forestReplaceChance;
            return this;
        }

        /** Sets the function for handling speed ups.
         *
         * @param speedUp a function.
         * @return the tile builder
         */
        public TileBuilder setSpeedUp(Function<Player, Integer> speedUp)
        {
            this.speedUp = speedUp;
            return this;
        }

        /** Marks the tile as non host for towns.
         *
         * @return the tile builder
         */
        public TileBuilder noTown()
        {
            this.canHostTown = false;
            return this;
        }
        
        /** Marks the tile as blocked for every movement.
         *
         * @return the tile builder
         */
        public TileBuilder noMovement()
        {
            this.blocksMovement = true;
            return this;
        }
        
        /** Sets the type of the tile.
         *
         * @param type a tile type
         * @return the tile builder
         */
        public TileBuilder setType(TileType type)
        {
            this.type = type;
            return this;
        }
        
        /** Marks the tile as host for paths.
         *
         * @return the tile builder
         */
        public TileBuilder pathHost()
        {
            this.canHostPath = true;
            return this;
        }
        
        /** Marks the tile as path.
         *
         * @return the tile builder
         */
        public TileBuilder path()
        {
            this.path = true;
            return this;
        }
        
        /** Sets the render type of the tile.
         *
         * @param type a render type
         * @return the tile builder
         */
        public TileBuilder setRenderType(TileRenderType type)
        {
            this.renderType = type;
            return this;
        }
        
        /** Returns the built tile.
         *
         * @return the built tile.
         */
        public Tile build()
        {
            return new Tile(energyCost, forestReplaceChance, speedUp, canHostTown, blocksMovement, type, canHostPath, path, renderType);
        }
    }
    
    private static int idCounter = 0;
    private static Tile[] tiles = new Tile[8];
    
    private static int addTile(Tile t)
    {
        if(idCounter >= tiles.length)
        {
            Tile[] newTiles = new Tile[tiles.length * 2];
            System.arraycopy(tiles, 0, newTiles, 0, tiles.length);
            tiles = newTiles;
        }
        tiles[idCounter] = t;
        return idCounter++;
    }
    
    /** Returns a tile from the given id or null if the id is invalid.
     *
     * @param id the id of the tile
     * @return a tile from the given id or null if the id is invalid
     */
    public static Tile fromId(int id)
    {
        if(id < 0 || id >= idCounter)
        {
            return null;
        }
        return tiles[id];
    }
    
    private final int id;
    private final int energyCost;
    private final float forestReplaceChance;
    private final Function<Player, Integer> speedUp;
    private final boolean canHostTown;
    private final boolean blocksMovement;
    private final TileType type;
    private final boolean canHostPath;
    private final boolean path;
    private final TileRenderType renderType;
    
    protected Tile(int energyCost, float forestReplaceChance, Function<Player, Integer> speedUp, 
            boolean canHostTown, boolean blocksMovement, TileType type, boolean canHostPath, boolean path,
            TileRenderType renderType)
    {
        id = addTile(this);
        this.energyCost = energyCost;
        this.forestReplaceChance = forestReplaceChance;
        this.speedUp = speedUp;
        this.canHostTown = canHostTown;
        this.blocksMovement = blocksMovement;
        this.type = type;
        this.canHostPath = canHostPath;
        this.path = path;
        this.renderType = renderType;
    }

    /** Returns the id of the tile.
     *
     * @return the id of the tile
     */
    public final int getId()
    {
        return id;
    }
    
    /** Returns the energy cost of a player on this tile.
     *
     * @param p a player
     * @return the energy cost of a player on this tile
     */
    public int getEnergyCost(Player p)
    {
        return energyCost - speedUp.apply(p);
    }

    /** Returns the chance that this tile is replaced by forest.
     *
     * @return the chance that this tile is replaced by forest
     */
    public float getForestReplaceChance()
    {
        return forestReplaceChance;
    } 
    
    /** Returns true if this tile can be replaced by a town.
     *
     * @return true if this tile can be replaced by a town
     */
    public boolean canHostTown()
    {
        return canHostTown;
    } 
    
    /** Called when the player fully enters a tile.
     *
     * @param p the player
     * @param map the current tilemap
     * @param x the x coordinate of the tile
     * @param y the y coordinate of the tile
     */
    public void onEnter(Player p, TileMap map, int x, int y)
    {
    }
    
    /** Called when the player leaves a tile.
     *
     * @param p the player
     * @param map the current tilemap
     * @param x the x coordinate of the tile
     * @param y the y coordinate of the tile
     */
    public void onLeave(Player p, TileMap map, int x, int y)
    {
    }

    /** Returns true if this tile blocks movement.
     *
     * @param p the player
     * @return true if this tile blocks movement
     */
    public boolean isBlockingMovement(Player p)
    {
        return blocksMovement || (p.isSailing() && type == TileType.LAND) || 
                (!p.isSailing() && type == TileType.DEEP_WATER);
    }
    
    /** Returns the type of the tile.
     *
     * @return the type of the tile
     */
    public TileType getType()
    {
        return type;
    }

    /** Returns true if this tile can be replaced by a path.
     *
     * @return true if this tile can be replaced by a path
     */
    public boolean canHostPath()
    {
        return canHostPath;
    }
    
    /** Returns true if this tile is a path.
     *
     * @return true if this tile is a path
     */
    public boolean isPath()
    {
        return path;
    }

    /** Returns the rendering type for overlays.
     *
     * @return the rendering type for overlays
     */
    public TileRenderType getRenderType()
    {
        return renderType;
    }
    
    /** Called when the player stands on that tile.
     *
     * @param p the player
     * @param map the current tilemap
     * @param x the x coordinate of the tile
     * @param y the y coordinate of the tile
     */
    public void isStandingOn(Player p, TileMap map, int x, int y)
    {
    }
}
